/**
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.distributedlog.exceptions;

import org.apache.distributedlog.thrift.service.ResponseHeader;
import org.apache.distributedlog.thrift.service.StatusCode;
import java.io.IOException;

/**
 * The basic exception of all distributedlog exceptions.
 */
public class DLException extends IOException {
    private static final long serialVersionUID = -4485775468586114393L;
    protected final StatusCode code;

    protected DLException(StatusCode code) {
        super();
        this.code = code;
    }

    protected DLException(StatusCode code, String msg) {
        super(msg);
        this.code = code;
    }

    protected DLException(StatusCode code, Throwable t) {
        super(t);
        this.code = code;
    }

    protected DLException(StatusCode code, String msg, Throwable t) {
        super(msg, t);
        this.code = code;
    }

    /**
     * Return the status code representing the exception.
     *
     * @return status code representing the exception.
     */
    public StatusCode getCode() {
        return code;
    }

    public static DLException of(ResponseHeader response) {
        String errMsg;
        switch (response.getCode()) {
            case FOUND:
                if (response.isSetErrMsg()) {
                    errMsg = response.getErrMsg();
                } else {
                    errMsg = "Request is redirected to " + response.getLocation();
                }
                return new OwnershipAcquireFailedException(errMsg, response.getLocation());
            case SUCCESS:
                throw new IllegalArgumentException("Can't instantiate an exception for success response.");
            default:
                if (response.isSetErrMsg()) {
                    errMsg = response.getErrMsg();
                } else {
                    errMsg = response.getCode().name();
                }
                return new DLException(response.getCode(), errMsg);
        }
    }
}
